#####
###  MailBoxer - a mailinglist/newsletter/mailarchive-framework for ZOPE
##   Copyright (C) 2003 Maik Jablonski (maik.jablonski@uni-bielefeld.de)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
##   along with this program; if not, write to the Free Software
###  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#####

# $Id: MailBoxer.py,v 1.28 2003/02/26 21:51:27 mjablonski Exp $

import StringIO, re, difflib, random, md5
import cgi, smtplib, rfc822, multifile, mimetools, mimify
from DateTime import DateTime
from Globals import InitializeClass, DTMLFile
from AccessControl import ClassSecurityInfo
from OFS.Folder import Folder
from zLOG import LOG, INFO
from MailBoxerTemplates import MailBoxerTemplates
import Bouncers

manage_addMailBoxerForm = DTMLFile('dtml/manage_addMailBoxerForm', globals())

def manage_addMailBoxer (self, id,
                title='', 
                mailto='', 
                moderator=[],
                moderated=0,
                archived='not archived',
                mtahosts=[],
                smtphost='127.0.0.1',
                REQUEST=None
            ):
    """ Add a new MailBoxer to current ObjectManager """

    mb = MailBoxer(id, title, mailto, moderator, moderated, archived, mtahosts)
    self._setObject(id,mb)
    mb = self._getOb(id)

    # Add a MailHost and a ZCatalog
    mb.manage_addFolder('archive','Archive')
    mb.manage_addProduct['MailHost'].manage_addMailHost('MailHost', 
                                                         smtp_host=smtphost)
    
    mb.manage_addProduct['ZCatalog'].manage_addZCatalog('Catalog','Catalog')
    mb.Catalog.manage_addIndex('mailDate', 'FieldIndex')
    mb.Catalog.manage_addIndex('mailFrom', 'TextIndex')
    mb.Catalog.manage_addIndex('mailSubject','TextIndex')
    mb.Catalog.manage_addIndex('mailBody', 'TextIndex')
    
    # Add dtml-templates
    for (id, data) in MailBoxerTemplates.items():
        mb.addDTMLMethod(id, file=data)
        
    # Redirect if requested TTW
    if REQUEST:
        return self.manage_main(self, REQUEST, update_menu=1)
        

class MailBoxer(Folder):
    """ A folder with maillist-managing-features. """

    security = ClassSecurityInfo()
    
    meta_type = 'Mail Boxer'

    manage_options = Folder.manage_options + ({'label': 'Bounce', 'action': 'manage_bounceForm'},)
    
    archive_options = ['not archived', 'plain text', 'with attachments']

    # The _hashkey is used for the md5-checksum for verifying (un)subscriptions. 
    # It will be reset every time you restart your Zope-Server.
    _hashkey   = str(random.random())
    
    # init properties
    title = ''
    mailto = ''
    maillist = []
    moderator = []
    moderated = 0
    plainmail = 0
    archived = ''
    subscribe = ''
    unsubscribe = ''
    mtahosts = []
    spamlist = []
    atmask = ''
    sniplist = ''
    catalog = ''
    xmailer = ''
    bounces = {}

    _properties=(
        {'id':'title', 'type':'string', 'mode':'w'},
        {'id':'mailto', 'type':'string', 'mode':'w'},
        {'id':'maillist', 'type':'lines', 'mode':'w'},
        {'id':'moderator', 'type':'lines', 'mode':'w'},
        {'id':'moderated', 'type':'boolean', 'mode':'w'},
        {'id':'plainmail', 'type':'boolean', 'mode':'w'},
        {'id':'archived', 'type':'selection','mode':'w', 
                      'select_variable':'archive_options'},
        {'id':'subscribe', 'type':'string', 'mode':'w'},
        {'id':'unsubscribe','type':'string', 'mode':'w'},
        {'id':'mtahosts', 'type':'tokens', 'mode':'w'},
        {'id':'spamlist', 'type':'lines', 'mode':'w'},
        {'id':'atmask', 'type':'string', 'mode':'w'},
        {'id':'sniplist', 'type':'lines', 'mode':'w'},
        {'id':'catalog', 'type':'string', 'mode':'w'},
        {'id':'xmailer', 'type':'string', 'mode':'w'}
    )


    def __init__(self, id, title, mailto, moderator, moderated, archived, mtahosts):
        """ setup a MailBoxer with reasonable defaults """
        
        self.id = id
        self.title = title
        self.mailto = mailto
        self.maillist = []
        self.moderator = moderator
        self.moderated = moderated
        self.plainmail = 0
        self.archived = archived
        self.subscribe = 'subscribe'
        self.unsubscribe = 'unsubscribe'
        self.mtahosts = mtahosts
        self.spamlist = []
        self.atmask = '(at)'
        self.sniplist = [r'(\n>[^>].*)+|(\n>>[^>].*)+', 
                         r'(?s)\n-- .*', 
                         r'(?s)\n_+\n.*']
        self.catalog = 'Catalog'
        self.xmailer = 'MailBoxer'

   
    # Public methods, callable TTW, protected by manage_checkboxer

    security.declarePublic('manage_mailboxer')
    def manage_mailboxer(self, REQUEST):
        """ Default for a all-in-one mailinglist-workflow. """

        # Check for correct ip, loops, spam...
        if self.manage_checkboxer(REQUEST):
            return
                    
        # Check for subscription/unsubscription
        if self.manage_requestboxer(REQUEST):
            return

        # Process the mail...
        if self.manage_frameboxer(REQUEST):
            return
            
            
    security.declarePublic('manage_frameboxer')
    def manage_frameboxer(self, REQUEST):
        """ The universal frame for a mailinglist. """

        if self.manage_checkboxer(REQUEST):
            return
         
        (mailHeader, mailBody) = self.splitMail(REQUEST['Mail'])
        
        From = self.mime_decode_header(mailHeader.get('from',''))
        (RealName, RealEmail) = self.parseaddr(From)
        
        # lower-case all addresses for comparisons
        LowEmail = RealEmail.lower()

        # Get lowered members from maillist-property
        MailList=[addr.lower() for addr in self.maillist]

        # Get additional members over mail_in_members-hook
        MailInMembers = []
        if hasattr(self, 'mail_in_members'):
            MailInMembers=[addr.lower() for addr in self.mail_in_members()]

        # Get moderators
        ModeratorList=[addr.lower() for addr in self.moderator]

        # message to a moderated list... relay all mails from a moderator
        if self.moderated:
            if LowEmail in ModeratorList:
                self.mail_handler(self, REQUEST, mail=mailHeader,
                                                        body=mailBody)
            else:
                self.mail_moderator(self, REQUEST, mail=mailHeader,
                                                        body=mailBody)
            return RealEmail

        # traffic! relay all mails from members and moderators...

        if LowEmail in MailList + MailInMembers + ModeratorList:
            self.mail_handler(self, REQUEST, mail=mailHeader, body=mailBody)
            return RealEmail

        # if all previous tests fail, it must be an unknown sender.
        self.mail_reply(self, REQUEST, mail=mailHeader, body=mailBody)
        return


    security.declarePublic('manage_requestboxer')
    def manage_requestboxer(self, REQUEST):
        """ Handles un-/subscribe-requests. """
        
        if self.manage_checkboxer(REQUEST):
            return
        
        (mailHeader, mailBody) = self.splitMail(REQUEST['Mail'])
        
        # get subject
        Subject = self.mime_decode_header(mailHeader.get('subject',''))

        # get email-address
        From = self.mime_decode_header(mailHeader.get('from',''))
        (RealName, RealEmail) = self.parseaddr(From)

        MailList=[addr.lower() for addr in self.maillist]
        
        # subscription? only subscribe if subscription is enabled.
        if (self.subscribe<>''
            and re.match('(?i)' + self.subscribe + "|.*: " + self.subscribe, Subject)):

            if RealEmail.lower() not in MailList:
                if Subject.find(self.pin(From))<>-1:
                    self.manage_addMember(RealEmail)
                self.mail_subscribe(self, REQUEST, mail=mailHeader, body=mailBody)
            else:
                self.mail_reply(self, REQUEST, mail=mailHeader, body=mailBody)

            return RealEmail

        # unsubscription? only unsubscribe if unsubscription is enabled...
        if (self.unsubscribe<>''
            and re.match('(?i)' + self.unsubscribe + "|.*: " + self.unsubscribe, Subject)):

            if RealEmail.lower() in MailList:
                if Subject.find(self.pin(From))<>-1:
                    self.manage_delMember(RealEmail)
                self.mail_unsubscribe(self, REQUEST, mail=mailHeader, body=mailBody)
            else:
                self.mail_reply(self, REQUEST, mail=mailHeader, body=mailBody)

            return RealEmail
                


    security.declarePublic('manage_listboxer')
    def manage_listboxer(self, REQUEST):
        """ Send a mail to all members of the list. """ 

        if self.manage_checkboxer(REQUEST):
            return
            
        Mail = REQUEST['Mail']
        
        # store mail in the archive? get context for the mail...
        context = self
        if self.archived <> 'not archived':
            context = self.manage_addMail(Mail)

        (mailHeader, mailBody) = self.splitMail(Mail)

        # plainmail-mode? get the plain-text of mail...
        if self.plainmail:          
            (TextBody, ContentType, HtmlBody) = self._unpackMultifile(None, 
                                                   multifile.MultiFile(
                                                    StringIO.StringIO(Mail)))
            if ContentType:
                mailHeader['content-type'] = ContentType
                mailBody = TextBody
            else:
                mailHeader['content-type'] = 'text/plain; charset=ISO-8859-1;'
                mailBody = self.HtmlToText(HtmlBody)
        
        # get custom header & footer
        (customHeader, customBody) = self.splitMail(self.mail_header(context,
                                                    REQUEST, title=self.title, 
                                                    mail=mailHeader, 
                                                    body=mailBody).strip())
                                                    
        customFooter = self.mail_footer(context, REQUEST, title=self.title,
                                                          mail=mailHeader, 
                                                          body=mailBody).strip()
            
        
        # Patch msg-headers with customHeaders
        msg = mimetools.Message(StringIO.StringIO(Mail))
        
        for hdr in customHeader.keys():
            if customHeader[hdr]:
                msg[hdr.capitalize()]=customHeader[hdr]
            else:
                del msg[hdr]
        
        newHeader = ''.join(msg.headers)

        # If customBody is not empty, use it as new mailBody
        if customBody.strip():
            mailBody=customBody

        newMail = "%s\r\n%s\r\n%s" % (newHeader, mailBody, customFooter)

        # Get additional Members from mail_out_members-hook
        MailOutMembers = self.maillist[:]
        if hasattr(self,'mail_out_members'):
            MailOutMembers = MailOutMembers + self.mail_out_members()
        
        # bulk mail. return-path is first moderator.

        try:
            smtpserver = smtplib.SMTP(self.MailHost.smtp_host, 
                                       int(self.MailHost.smtp_port))
            smtpserver.sendmail(self.moderator[0], MailOutMembers, newMail)
            smtpserver.quit()
        except:
            LOG('MailBoxer', INFO, 'SMTP-Error')


    security.declarePublic('manage_inboxer')
    def manage_inboxer(self, REQUEST):
        """ Wrapper to mail directly into archive. """

        if self.manage_checkboxer(REQUEST):
            return
            
        self.manage_addMail(REQUEST['Mail'])


    security.declarePublic('manage_bounceboxer')
    def manage_bounceboxer(self, REQUEST):
        """ Check for bounced mails. """

        # Check for bounce; to prevent raising exceptions for wired mails,
        # catch all...
                
        try:
            bouncedAddresses = Bouncers.ScanMessage(REQUEST['Mail'])
            if bouncedAddresses:
                mail_out_members=[]
                if hasattr(self,'mail_out_members'):
                    mail_out_members = self.mail_out_members()

                # Get lowered member-list
                MailList = [addr.lower() for addr in self.maillist +
                                                     mail_out_members]

                for item in bouncedAddresses:
                    if item.lower() in MailList:
                        # Create an entry for bouncer, if it not exists...
                        if not self.bounces.has_key(item):
                            self.bounces[item]=[0, DateTime(), DateTime()]

                        # Count up and remember last bounce
                        self.bounces[item] = [self.bounces[item][0]+1,
                                              self.bounces[item][1], DateTime()]

                        self.bounces = self.bounces

                message = 'Mail bounced for: %s' % ', '.join(bouncedAddresses)
                LOG('MailBoxer', INFO, message)
                return bouncedAddresses

        except:
            return None
            

    security.declarePublic('manage_checkboxer')
    def manage_checkboxer(self, REQUEST):
        """ Check for ip, bounce, loop and spam. """
        
        # Check for correct IP        
        if self.mtahosts:
            if 'HTTP_X_FORWARDED_FOR' in self.REQUEST.environ.keys():
                REMOTE_IP = self.REQUEST.environ['HTTP_X_FORWARDED_FOR']
            else:
                REMOTE_IP = self.REQUEST.environ['REMOTE_ADDR']

            if REMOTE_IP not in self.mtahosts:          
                message = 'Host %s is not allowed' % (REMOTE_IP)
                LOG('MailBoxer', INFO,  message)
                return message

        # Check for loop
        (mailHeader,mailBody)=self.splitMail(REQUEST['Mail'])

        if mailHeader.get('x-mailer')==self.xmailer:
            message = 'Mail already processed'
            LOG('MailBoxer', INFO, message)
            return(message)

        # Check for spam
        for regexp in self.spamlist:
            if regexp and re.search(regexp,REQUEST['Mail']):
                message='Spam detected: %s\n\n%s\n' % (regexp,REQUEST['Mail'])
                LOG('MailBoxer', INFO, message)
                return message

        return

            
    # Protected methods
    
    security.declareProtected('Manage properties','manage_addMember')
    def manage_addMember(self, email):
        """ Add member to maillist. """

        MailList=[addr.lower() for addr in self.maillist]
        
        if email.lower() not in MailList:
            self.maillist.append(email)
            self.maillist.sort()
            self.maillist = self.maillist
            return email
            

    security.declareProtected('Manage properties','manage_delMember')
    def manage_delMember(self, email):
        """ Deletes member from maillist. """

        MailList=[addr.lower() for addr in self.maillist]
        
        if email.lower() in MailList:
            index = MailList.index(email.lower())
            self.maillist = self.maillist[:index] + self.maillist[index+1:]
            return email
            

    security.declareProtected('Add Folders','manage_addMail')
    def manage_addMail(self, Mail):
        """ store mail & attachments in a folder and return it """

        (mailHeader, mailBody) = self.splitMail(Mail)

        # always take our own date... don't believe the clients!
        time = DateTime()
        
        # now let's create the date-path (yyyy/yyyy-mm) 
        year  = str(time.year())                  # yyyy
        month = "%s-%s" % (year, str(time.mm()))  # yyyy-mm

        archive = getattr(self,'archive')
        
        # do we have a year folder already?
        if not hasattr(archive, year):
            archive.manage_addFolder(year, title=year)
        yearFolder=getattr(archive, year)

        # do we have a month folder already?
        if not hasattr(yearFolder, month):
            yearFolder.manage_addFolder(month, title=month)
        monthFolder=getattr(yearFolder, month)

        # let's create the mailObject
        mailFolder = monthFolder
        
        Subject = self.mime_decode_header(mailHeader.get('subject',
                                                         'No Subject'))
        From = self.mime_decode_header(mailHeader.get('from','No From'))
        Title = "%s / %s" % (Subject, From)

        # maybe it's a reply ?
        if ':' in Subject:
            for currentFolder in monthFolder.objectValues(['Folder']):
                if difflib.get_close_matches(Subject,
                                             [currentFolder.mailSubject]):
                    mailFolder=currentFolder

        # search a free id for the mailobject
        id = time.millis()
        while hasattr(mailFolder,str(id)):  
             id = id + 1
             
        id = str(id)

        mailFolder.manage_addFolder(id, title=Title)
        mailObject = getattr(mailFolder, id)

        # unpack attachments
        (TextBody, ContentType, HtmlBody) =  self._unpackMultifile(mailObject, 
                                                     multifile.MultiFile(
                                                      StringIO.StringIO(Mail)))

        # ContentType is only set for the TextBody
        if ContentType:
            mailBody = TextBody
        else:
            mailBody = self.HtmlToText(HtmlBody)
        
        # and now add some properties to our new mailobject 
        mailObject.manage_addProperty('mailFrom', From, 'string')
        mailObject.manage_addProperty('mailSubject', Subject, 'string')
        mailObject.manage_addProperty('mailDate', time, 'date')
        mailObject.manage_addProperty('mailBody', mailBody, 'text')
        
        # Index the new created mailFolder in the catalog
        if hasattr(self, self.catalog):
            getattr(self, self.catalog).catalog_object(mailObject)
    
        return mailObject


    security.declareProtected('Manage properties','manage_bounceForm')
    manage_bounceForm = DTMLFile('dtml/manage_bounceForm', globals())


    security.declareProtected('Manage properties','manage_resetBounces')
    def manage_resetBounces(self, ids, REQUEST=None):
        """ Removes bounced addresses from bounces-list. """

        for addr in ids:
            if self.bounces.has_key(addr):
                del(self.bounces[addr])
        
        self.bounces = self.bounces
        
        if REQUEST:
            return self.manage_bounceForm(REQUEST, update_menu=1)
            

    security.declareProtected('Manage properties','manage_deleteBounces')
    def manage_deleteBounces(self, ids, REQUEST=None):
        """ Deletes bounced addresses from maillist. """

        for addr in ids:
            self.manage_delMember(addr)

        self.manage_resetBounces(ids)
                
        if REQUEST:
            return self.manage_bounceForm(REQUEST, update_menu=1)
   

    # Public methods, but not callable TTW

        
    def splitMail(self, Mail):
        # return (header,body) of a mail given as string

        msg = mimetools.Message(StringIO.StringIO(Mail))
        
        # Get headers
        mailHeader = {}
        for (key,value) in msg.items():
            mailHeader[key] = value
        
        # Get body
        msg.rewindbody()
        mailBody = msg.fp.read()

        return (mailHeader, mailBody)

          
    def pin(self, mailFrom):
        # returns the hex-digits for a randomized md5.
        
        (RealName, RealEmail) = self.parseaddr(mailFrom)      
        return md5.new(RealEmail.lower() + self._hashkey).hexdigest()
        

    def mime_decode_header(self, strg):
        # Returns the undecoded header
        
        return mimify.mime_decode_header(strg)
        

    def parseaddr(self, header):
        # wrapper for rfc822.parseaddr, returns (name, addr)
        
        return rfc822.parseaddr(header)
    
    
    def snipMail(self, strg):
        # snips all regexps from strg and urlifies it
        
        for regexp in self.sniplist:
            if regexp: strg = re.sub(regexp,'[...]',strg)

        strg = cgi.escape(strg)
        
        # Urlifing can be done better... but it's quite good
        strg = re.sub(r'(?P<url>http://[-_&;,?:~=%#+/.0-9a-zA-Z]+)', 
                      r'<a href="\g<url>">\g<url></a>', strg)

        return strg.strip()


    def at(self, strg):
        # replaces all @ with the atmask 
        
        return(strg.replace('@',self.atmask))
        


    def nextObject(self, context):
        # Returns next Folder in context or None
           
        parentObjects = context.aq_parent.objectIds('Folder')
        parentObjects.sort()

        if context.getId() in parentObjects:
            parentPosition = parentObjects.index(context.getId())
            
            if parentPosition+1<len(parentObjects):
                return getattr(context,str(parentObjects[parentPosition+1]))
               
            elif context<>self and context.aq_parent<>self:
                if self.nextObject(context.aq_parent):
                    objList = self.nextObject(context.aq_parent).objectValues('Folder')
                    if objList:
                        return objList[0]

        return None


    def previousObject(self, context):
        # Return previous Folder in context or None
           
        parentObjects = context.aq_parent.objectIds('Folder')
        parentObjects.sort()
 
        if context.getId() in parentObjects:
            parentPosition = parentObjects.index(context.getId())
                   
            if parentPosition > 0:
                return getattr(context,str(parentObjects[parentPosition-1]))
               
            elif context<>self and context.aq_parent<>self:
                if self.previousObject(context.aq_parent):
                    objList = self.previousObject(context.aq_parent).objectValues('Folder')
                    if objList:
                        return objList[-1]
               
        return None


    def HtmlToText(self, html):
        # converts html to text
        
        p = HTMLStripper()
        p.feed(html)
        p.close()
        return str(p)
        

    # Private methods
    
    security.declarePrivate('_unpackMultifile')
    def _unpackMultifile(self, obj, multifile):
        """ Unpack body and attachments; return (TextBody, ContentType, HtmlBody). """

        TextBody = HtmlBody = ContentType = ''
        
        msg = mimetools.Message(multifile)
        maintype = msg.getmaintype()
        subtype = msg.getsubtype()
        name = msg.getparam('name')

        # Recurse over all nested multiparts
        if maintype == 'multipart':
            multifile.push(msg.getparam('boundary'))
            multifile.readlines()
            while not multifile.last:
                multifile.next()
                
                (tmpTextBody, tmpContentType, tmpHtmlBody) = \
                                           self._unpackMultifile(obj, multifile)
                
                # Return ContentType only for the plain-body of a mail
                if tmpContentType:
                    TextBody = tmpTextBody
                    ContentType = tmpContentType
                
                if tmpHtmlBody:
                    HtmlBody = tmpHtmlBody
                    
            multifile.pop()
            return (TextBody,ContentType,HtmlBody)

        # Process MIME-encoded data
        plainfile = StringIO.StringIO()

        try:
            mimetools.decode(multifile,plainfile,msg.getencoding())
        # unknown or no encoding? 7bit, 8bit or whatever... copy literal
        except ValueError:
            mimetools.copyliteral(multifile,plainfile)

        body = plainfile.getvalue()
        plainfile.close()
    
        # Get plain text
        if maintype == 'text' and subtype == 'plain' and not name:
            TextBody = body
            ContentType = msg.get('content-type', 'text/plain')
            
        # Decode the rest
        else:
            # No name? This should be the html-body...
            if not name:
                name = '%s.%s' % (maintype,subtype)
                HtmlBody = body

            if obj and self.archived=='with attachments':
           
                id = DateTime().millis()
                # to be sure: test and search for a free id...
                while hasattr(obj,str("%s.%s" % (str(id), subtype))):  
                    id = id + 1
                    
                id = "%s.%s" % (str(id), subtype)
                obj.manage_addFile(id, title = self.mime_decode_header(name),
                                       file = body,
                                       content_type = maintype + '/' + subtype)

        return (TextBody, ContentType, HtmlBody)


# And here we go to the moon... Houston, we have a problem...
InitializeClass(MailBoxer)


# HTMLStripper is inspired by StripTagParser from Dieter Maurer

from sgmllib import SGMLParser
class HTMLStripper(SGMLParser):
    """ Remove tags and translate entities. """

    from htmlentitydefs import entitydefs

    text = ''

    def handle_data(self, data):
        self.text = self.text + data

    def __str__(self):
        return self.text.strip()
